/* This file is part of an "unsupported" FreeRTOS port distribution for
 * Freescale MC9S12DP256B CPU.
 */

/* Scheduler include files. */
#include "projdefs.h"
#include "portable.h"
#include "task.h"
#include "version.h"

volatile unsigned char vPortSchRunning = 0;
volatile unsigned char vPortCriticalNesting = 0;
volatile unsigned char vPortTmp = 0;

extern volatile void *pxCurrentTCB;

/* variable for temporary variables storage, placed in reserved memory block */
unsigned char vPortCCRSave;
unsigned short vPortAddrSave;
unsigned char vPortPPAGESave;

/* timer setup routine */
void prvSetupTimerInterrupt(void);

/*
** Tick ISR
*/
#pragma interrupt_handler rtiHandler()
void rtiHandler (void)
{
   vPortIsrHead();

   /* do interrupt stuff... */

	vPortIsrTail();
}

/*
 * See header file for description.
 */
portSTACK_TYPE *pxPortInitialiseStack(portSTACK_TYPE *pxTopOfStack,
	pdTASK_CODE pxCode, void /*@unused@*/ *pvParameters)
{
	/* Place a few bytes of known values on the bottom of the stack.
	 * This is just useful for debugging. */
	*pxTopOfStack = (portSTACK_TYPE) 0x11;
	pxTopOfStack--;
	*pxTopOfStack = (portSTACK_TYPE) 0x22;
	pxTopOfStack--;
	*pxTopOfStack = (portSTACK_TYPE) 0x33;
	pxTopOfStack--;

	/* Simulate how the stack would look after a call to vPortYield() generated
	 * by the compiler.
	 *
	 * First store the function parameters.  This is where the task will expect
	 * to find them when it starts running.
	 * A small note: CodeWarrior compiler passes single pointer via D and B
	 * registers, no stack is used, no need to push parameters into stack, but...
	 * there's also a problem to pass those parameters to the task, right now
	 * parameters passing to function is not supported. */
//	*pxTopOfStack = (portSTACK_TYPE) ((unsigned int) pvParameters);
//	pxTopOfStack--;
//	*pxTopOfStack = (portSTACK_TYPE) ((unsigned int) pvParameters >> 8);
//	pxTopOfStack--;

	/* here comes the return address, since interrupts don't use pages, only
	 * two bytes of return address are stored, PPAGE byte will be saved later,
	 * manually
	 * such a complex lines are compiled into two assembly commands each */
	*pxTopOfStack = (portSTACK_TYPE) *(((portSTACK_TYPE *) (&pxCode)) + 1);
	pxTopOfStack--;
	*pxTopOfStack = (portSTACK_TYPE) *(((portSTACK_TYPE *) (&pxCode)) + 0);
	pxTopOfStack--;

	/* Next are all the registers that form part of the task context. */

	/* Y register */
	*pxTopOfStack = (portSTACK_TYPE) 0xff;
	pxTopOfStack--;
	*pxTopOfStack = (portSTACK_TYPE) 0xee;
	pxTopOfStack--;

	/* X register */
	*pxTopOfStack = (portSTACK_TYPE) 0xdd;
	pxTopOfStack--;
	*pxTopOfStack = (portSTACK_TYPE) 0xcc;
	pxTopOfStack--;

	/* A register */
	*pxTopOfStack = (portSTACK_TYPE) 0xbb;
	pxTopOfStack--;

	/* B register */
	*pxTopOfStack = (portSTACK_TYPE) 0xaa;
	pxTopOfStack--;

	/* CCR: pay attention that on task call, interrupts will be reenabled since
	 * "I" bit of CCR is cleared */
	*pxTopOfStack = (portSTACK_TYPE) 0x00;
	pxTopOfStack--;

	/* save the page of current task (will be resored manually inside the
	 * ISR) */
   *pxTopOfStack = (portSTACK_TYPE) ((int) pxCode);

	return pxTopOfStack;
}

/*-----------------------------------------------------------*/
portSHORT sPortStartScheduler( portSHORT sUsePreemption )
{
	if (sUsePreemption != 0)
	{
		/* setup the timer to generate timebase interrupt */
		prvSetupTimerInterrupt();
	}

   /* flag the scheduler as started so vPortYield() knows it */
   vPortSchRunning = 1;

	/* Restore the context of the first task to run. */
	portRESTORE_CONTEXT();

	/* simulate a function call end as generaged by the compiler. We will now
	 * jump to the start of the task the context of which we have just restored
	 */
	__asm__("rti");

	/* should not get here, done to avoid compiler warning */
	return pdTRUE;
}

/*-----------------------------------------------------------*/

void vPortEndScheduler( void )
{
	/* It is unlikely that the scheduler for the Motorola port will get stopped
	 * once running.  If required disable the tick interrupt here, then return
	 * to sPortStartScheduler(). */
}

/*-----------------------------------------------------------*/

/** Manual context switch. */
void volatile vPortYield(void)
{
   /* Avoid yield if we are not running the RTOS. Usefull if you
    * want to use a common set of code and functions both as tasks
    * and globals. */
   if (vPortSchRunning == 0)
   {
     return;
   }

	/* here we should save all the registers as if this was done by interrupt
	 * call */

	/* no context switch should occur during manual contex switch */
   portDISABLE_RTI_INTERRUPT();

	/* simulate hardware interrupt handler by saving approriate registers
	 * but... keep interrupt enable bit clear, so on activation of this task
	 * interrupts will be reenabled. */
	__asm__("pshy");
	__asm__("pshx");
	__asm__("psha");
	__asm__("pshb");
	__asm__("pshc");
	__asm__("bclr 0, sp, 0x10");

	/* complete context switch save */
	portSAVE_CONTEXT();

	/* Switch to the highest priority task that is ready to run. */
	vTaskSwitchContext();

	/* Start executing the task we have just switched to. */
	portRESTORE_CONTEXT();

	/* restore tasks's page */
	PPAGE = vPortPPAGESave;

   /* reenable RTI interrupt, in order to disallow reentry, disable interrupts
    * globally beforehand */
   portDISABLE_INTERRUPTS();
   portENABLE_RTI_INTERRUPT();
	/* don't enable interrupts, RTI will worry for that */
	__asm__("rti");
}


/** setup timebase interrupt */
#define RTIF   1<<7   /* Real Time Interrupt Flag */
#define RTIE   1<<7   /* Real Time Interrupt bit */
void prvSetupTimerInterrupt(void)
{
   /*
   ** Enable the RTI interrupt at 1.024ms
   ** OSCCLK is external oscillator
   ** OSCCLK = 8MHz assumed ->  1ms = 1.024 / ( OSCCLK / ( 8 * 2^10 ) )
   */
   RTICTL  = 0x17;   /* RTR6:4 = 1, RTR3:0 = 7 => 8*2^10 */
   CRGFLG |= RTIF;   /* clear RTI flag */
   CRGINT |= RTIE;   /* enable RTI interrupt */
}

/*-----------------------------------------------------------*/
/** @brief save CCR value into global varible: no stack is used.
 * We have to count nesting of critical sections enterances and save interrupts
 * state only if this is the first level, nesting counter is protected by
 * semaphore */
void vPortEnterCritical(void)
{
   __asm__("tpa");
   __asm__("staa _vPortTmp");
   portDISABLE_INTERRUPTS();

   /* no interrupts are allowed here, so nobody is able to access
    * "vPortCriticalNesting", no need to protect it */
   if (vPortCriticalNesting == 0)
   {
      vPortCCRSave = vPortTmp;
   }
   vPortCriticalNesting++;
}

/** restore CCR from global variable: no stack is used */
void vPortExitCritical(void)
{
   vPortCriticalNesting--;

   /* check this is the first level of "enter critical" section and if that is
    * the case, clear the interrupt flag if needed */
   if (/*vPortCriticalNesting == 0 && */ (vPortCCRSave & 0x10) == 0)
   {
      /* reenable interrupts */
      portENABLE_INTERRUPTS();
   }
}
